/* an example of calling the NAG Sobol generator nag_quasi_rand_uniform (g05ymc)
   with scrambled sequences in an OpenMP multi-threaded environment */
#include <stdio.h>
#include <math.h>
#include <nag.h>
#include <nagx04.h>
#include <nag_stdlib.h>
#include <nagg05.h>
#include <omp.h>

int main(int argc, char *argv[]) {
  int i, exit_status = 0;
  Integer NumberOfThreads, liref, pdquas, iskip, tskip, ThreadNo, npaths, idim,
    tnpaths, t1, lstate;
  double sumRandom;
  Integer *iref = 0, *state = 0;
  Integer seed[1];
  double *quas = 0;  
  NagError fail;

  /* we are currently just summing the values for the first dimension, so makes 
     sense to work in ColMajor order, where all values from a given dimension 
     are contiguous */
  Nag_OrderType order = Nag_ColMajor;

  /* want the Sobol generator and Owen-like scrambling */
  Nag_QuasiRandom_Sequence genid = Nag_QuasiRandom_Sobol;
  Nag_QuasiRandom_Scrambling stype = Nag_OwenLike;

  /* scramble the default number of digits */
  Integer nsdigi = 0;

  /* use the Mersenne Twister psuedo-generator */
  Nag_BaseRNG pgenid = Nag_MersenneTwister;
  Integer psubid = 0;

  /* only using one seed value */
  Integer lseed = 1;

  /* parse the input arguments */
  if (argc < 4) {
    printf("Usage: XXX n d i s\n");
    printf("Where:\n");
    printf("  n = number of paths\n");
    printf("  d = number of dimensions\n");
    printf("  i = initial skip\n");
    printf("  s = seed (optional)\n");
    return 0;
  }

  /* get the problem size */
  npaths = atoi(argv[1]);
  idim = atoi(argv[2]);
  iskip = atoi(argv[3]);
  if (argc > 4) {
    /* use the supplied value for the seed */
    seed[0] = atoi(argv[4]);
  } else {
    /* use a default value for the seed */
    seed[0] = 1762543;
  }

  /* intialise the error structure */
  INIT_FAIL(fail);

  /* get the length of the state array for the psuedo-generator */
  lstate = -1;
  nag_rand_init_repeatable(pgenid, psubid, seed, lseed, state, &lstate, &fail);
  if (fail.code != NE_NOERROR)
    {
      printf("Error from nag_rand_init_repeatable (g05kfc).\n%s\n", 
             fail.message);
      exit_status = 1;
      goto END;
    }

  /* calculate the size of the reference vector */
  liref = 32 * idim + 7;

#pragma omp parallel default(none) \
  reduction(+:sumRandom) \
    private(ThreadNo,tskip,tnpaths,quas,iref,fail,i,t1,pdquas,\
            NumberOfThreads,state) \
      shared(npaths,exit_status,idim,liref,iskip,genid,order,pgenid,\
             psubid,lstate,seed,lseed,nsdigi,stype)
  {

    /* intialise the error structure */
    INIT_FAIL(fail);

    /* thread number */
    ThreadNo = omp_get_thread_num();

    /* number of threads */
    NumberOfThreads = omp_get_num_threads();

    /* calculate the skip and the number of values to generate for this thread */
    if (NumberOfThreads == 1) {
      tskip = 0;
      tnpaths = npaths;
    } else {
      t1 = npaths % NumberOfThreads;
      t1 = (ThreadNo < t1) ? ThreadNo : t1;
      tskip = ThreadNo * (npaths / NumberOfThreads) + t1;
      tnpaths = (npaths + NumberOfThreads - ThreadNo - 1) / NumberOfThreads;
    } 

    /* adjust the skip for this thread by the initial skip */
    tskip += iskip;

    /* calculate the stride */
    pdquas = (order == Nag_RowMajor)?idim:tnpaths;

    /* allocate memory to hold the generated values (quas), state of the 
       quasi-random generator (iref) and state of the pseudo generator (state) */
    if (!(quas = NAG_ALLOC(tnpaths*idim, double)) ||
        !(state = NAG_ALLOC(lstate,Integer)) ||
        !(iref = NAG_ALLOC(liref, Integer)))
      {
        printf("Allocation failure in thread %"NAG_IFMT"\n", ThreadNo + 1);
        exit_status = -1;
        goto TEND;
      }
    
    /* nag_rand_init_repeatable (g05kfc): initialise the pseudo generator, all 
       threads need to be initialised using the same seed as we want to use the
       same scrambling on each thread, each thread needs its own copy of state 
       as this is input / output to NAG routine */
    nag_rand_init_repeatable(pgenid, psubid, seed, lseed, state, &lstate, &fail);
    if (fail.code != NE_NOERROR)
      {
        printf("Error from nag_rand_init_repeatable (g05kfc) in thread "
               "%"NAG_IFMT".\n%s\n", ThreadNo + 1, fail.message);
        exit_status = 1;
        goto TEND;
      }

    /* nag_quasi_init_scrambled (g05ync): initialise the quasi generator */
    nag_quasi_init_scrambled(genid, stype, idim, iref, liref, tskip, nsdigi,
                             state, &fail);
    if (fail.code != NE_NOERROR)
      {
        printf("Error from nag_quasi_init_scrambled (g05ync) in thread "
               "%"NAG_IFMT".\n%s\n",ThreadNo + 1, fail.message);
        exit_status = 1;
        goto TEND;
      }
    
    /* nag_quasi_rand_uniform (g05ymc): generate the required values */
    nag_quasi_rand_uniform(order, tnpaths, quas, pdquas, iref, &fail);
    if (fail.code != NE_NOERROR)
      {
        printf("Error from nag_quasi_rand_uniform (g05ymc) in thread "
               "%"NAG_IFMT".\n%s\n",ThreadNo + 1, fail.message);
        exit_status = 1;
        goto TEND;
      }
    
    /* calculate the sum for this particular thread. We are only interested in 
       the first dimension which are stored in the first tnpaths of the quas
       array */
    for (i = 0; i < tnpaths; i++)
      sumRandom += quas[i];

    printf("Thread = %4"NAG_IFMT", Sum = %g\n", ThreadNo+1,sumRandom);
    
    /* free memory */
  TEND:
    if (quas) free(quas);
    if (iref) free(iref);
    if (state) free(state);
  }
  
  /* display the results */
  if (!exit_status) {
    printf("\nTotal Sum = %g\n", sumRandom);
  }

 END:
  return exit_status;
}
